Affected by GO-2025-4237
and 1 other vulnerabilities
GO-2025-4237: Weaviate OSS has a Path Traversal Vulnerability via Backup ZipSlip in github.com/weaviate/weaviate
GO-2025-4238: Weaviate OSS has path traversal vulnerability via the Shard Movement API in github.com/weaviate/weaviate
package
Version:
v1.31.7
Opens a new window with list of versions in this module.
Published: Jul 23, 2025
License: BSD-3-Clause
Opens a new window with license information.
Imports: 14
Opens a new window with list of imports.
Imported by: 0
Opens a new window with list of known importers.
Documentation
¶
type DocumentScore struct {
Document string `json:"document"`
Score float64 `json:"score"`
}
type RankInput struct {
Query string `json:"query"`
Documents []string `json:"documents"`
RankPropertyValue string `json:"property"`
}
type RankResponse struct {
Query string `json:"query"`
Scores []DocumentScore `json:"scores"`
RankPropertyValue string `json:"property"`
Score float64 `json:"score"`
Error string `json:"error"`
}
Source Files
¶
Click to show internal directories.
Click to hide internal directories.